#ifndef FILE_MYLIB
#define FILE_MYLIB

#ifdef __INTEL_COMPILER
#pragma message(" ##########  Use of MKL  ###############")
#include <mkl.h>
//#include <mkl_boost_ublas_matrix_prod.hpp>
// MKL
#else
#pragma message(" ##########  Use of CBLAS  ###############")
extern "C"
{
#include <cblas.h>               // cBLAS Library
#include <lapacke.h>             // Lapack
}
#endif

#include <cmath>           // exp()
#include <iostream>
#include <vector>

/** 	Dense matrix, rowise stored as 1D vector.
 */
class DenseMatrix {
public:
    /** 	Constructor with implicit initialization
     *
     * 	@param[in] nrow  number of rows
     * 	@param[in] mcol  number of columns
     *
     */
    DenseMatrix(int nrow, int mcol);

    /** 	Constructor
     *
     * 	@param[in] nrow  number of rows
     * 	@param[in] mcol  number of columns
     *  @param[in] val   initial value of matrix elements
     *
     */
    DenseMatrix(int nrow, int mcol, double val);

    DenseMatrix(DenseMatrix const &) = default;
    DenseMatrix(DenseMatrix &&) = default;
    DenseMatrix &operator=(DenseMatrix const &) = default;
    DenseMatrix &operator=(DenseMatrix &&) = default;

    /** 	Matrix times vector multiplication with rowise access.
     *
     * 	@param[in] u  vector
     * 	@return resulting vector M*u
     *
     */
    std::vector<double> Mult(std::vector<double> const &u) const;

    /** 	transposed Matrix times vector multiplication (columnwise) access.
     *
     * 	@param[in] u  vector
     * 	@return resulting vector @f$ M^T*u @f$
     *
     */
    std::vector<double> MultT(std::vector<double> const &u) const;

    /** 	transposed Matrix times vector multiplication (columnwise) access.
     *
     * 	@param[in] u  vector
     * 	@return resulting vector @f$ M^T*u @f$
     *
     */
    std::vector<double> MultT_fast(std::vector<double> const &u) const;


    /** 	Matrix-matrix multiplication (rowise) access.
     *
     * 	@param[in] B  matrix
     * 	@return resulting matrix @f$ A*B @f$
     *
     */
    DenseMatrix Mult(DenseMatrix const &B) const;


    /** 	Matrix-matrix multiplication (rowise) access.
     *
     * 	@param[in] B  matrix
     * 	@return resulting matrix @f$ A*B @f$
     *
     */
    DenseMatrix Mult_fast(DenseMatrix const &B) const;

    /** 	Matrix-matrix multiplication (rowise) access.
     *
     * 	@param[in] B  matrix
     * 	@return resulting matrix @f$ A*B @f$
     *
     */
    DenseMatrix Mult_Blas(DenseMatrix const &B) const;
    DenseMatrix Mult_cuBlas(DenseMatrix const &B) const;
    //DenseMatrix Mult_cuBlas_row(DenseMatrix const &B) const;  // doesn't work
    DenseMatrix Mult_cuBlas_Test(DenseMatrix const &B, int loopcount, double &seconds) const;


    /** 	Compares two matrices for equality.
     *
     * 	@param[in] B  matrix
     * 	@return equal regarding internal relative accuracy?
     *
     */
    bool operator==(DenseMatrix const &B) const;

    /** 	Get the number of rows.
    * 	@return number of rows.
    */
    int GetNrows() const
    {
        return _nrows;
    }
    /** 	Get the number of rows.
    * 	@return number of rows.
    */

    int GetNcols() const
    {
        return _mcols;
    }

    /** 	Access to element @p i in rowise numbering
    * 	@return value.
    */
    double &operator[] (int i)
    {
        return _A[i];
    }

    /** 	Access to element @p i in rowise numbering
    * 	@return value.
    */
    double const &operator[] (int i) const
    {
        return _A[i];
    }

    /**
     *  @return raw pointer to begin of element data.
    */
    double *data()
    {
        return _A.data();
    }

    /**
     *  @return raw pointer to begin of element data.
    */
    double const *data() const
    {
        return _A.data();
    }

private:
    int _nrows;       //!< number of rows in matrix
    int _mcols;       //!< number of coluns in matrix

protected:
    std::vector<double> _A;  //!< matrix values

};

/** 	Operator overloading: Matrix-vector product with the multiplication operator
 * 	@param[in] A	dense matrix (1D access)
 *  @param[in] u	vector
 *
 *	@return    resulting vector
*/
inline
std::vector<double> operator*(DenseMatrix const &A, std::vector<double> const &u)
{
    return A.Mult(u);
}
//----------------------------------------------------------------------


class Blas_DenseMatrix : public DenseMatrix {
public:
    /** 	Constructor
     *
     * 	@param[in] nrow  number of rows
     * 	@param[in] mcol  number of columns
     *
     */
    Blas_DenseMatrix(int nrow, int mcol);

    /** 	Constructor
     *
     * 	@param[in] nrow  number of rows
     * 	@param[in] mcol  number of columns
     *  @param[in] val   initial value of matrix elements
     *
     */
    Blas_DenseMatrix(int nrow, int mcol, double val);

    /** 	Matrix times vector multiplication with rowise access.
     *
     * 	@param[in] u  vector
     * 	@return resulting vector M*u
     *
     */
    std::vector<double> Mult(std::vector<double> const &u) const;

    /** 	transposed Matrix times vector multiplication (columnwise) access.
     *
     * 	@param[in] u  vector
     * 	@return resulting vector @f$ M^T*u @f$
     *
     */
    std::vector<double> MultT(std::vector<double> const &u) const;


    /** 	Matrix-matrix multiplication (rowise) access.
     *
     * 	@param[in] B  matrix
     * 	@return resulting matrix @f$ A*B @f$
     *
     */
    Blas_DenseMatrix Mult(Blas_DenseMatrix const &B) const;

};



//----------------------------------------------------------------------

/** Output operator for vector
 * 	@param[in,out] s	output stream, e.g. @p cout
 *  @param[in]     v    vector
 *
 *	@return    output stream
*/
template <class T>
std::ostream &operator<<(std::ostream &s, std::vector<T> const &v)
{
    for (auto vp : v) {
        s << vp << "  ";
    }
    return s;
}


///** Generates an equistant vector having @p n elements from interval [@p a, @p b].
 //*
 //* 	@param[in] a	interval start
 //*  @param[in] b    interval end
 //*  @param[in] n    number of equidistant points in interval
 //*
 //*	@return    vector with the equistant elements
//*/
//std::vector<double> linspace(double a, double b, int n = 100);


///** <a href="https://en.wikipedia.org/wiki/Sigmoid_function">Sigmoid</a> function.
 //*
 //* 	@param[in] x  evaluation point
 //*
 //*	@return    sigmoid value
//*/
//template <class T>
//T sigmoid(T x)
//{
    //return T(1.0) / (T(1.0) + std::exp(x));
//}



//----------------------------------------------------------------------

/** 	Dense matrix with a tensor product structure @f$ M = u^T * v @f$.
 */
class TensorMatrix {
public:
    /** 	Constructor of tensor product matrix for @f$ M = u^T * v @f$.
     *
     * 	@param[in] ucol  column vector
     * 	@param[in] vrow  row vector
     *
     */
    TensorMatrix(std::vector<double> const &ucol, std::vector<double> const &vrow);

    /** 	Matrix times vector multiplication.
     *
     *   Implementation realizes as @f$ f := M * x = u^T * (v * x) @f$
     *
     * 	@param[in] x  column vector
     * 	@return resulting vector M*x
     *
     */
    std::vector<double> Mult(std::vector<double> const &x) const;

    /** 	Transposed Matrix times vector multiplication.
     *
     *      Implementation realizes as @f$ f := M^T * y = v^T * (u * y) @f$.
     *
     * 	@param[in] y  vector
     * 	@return resulting vector @f$ M^T*y @f$
     *
     */
    std::vector<double> MultT(std::vector<double> const &y) const;

    /** 	Get the number of rows.
    * 	@return number of rows.
    */
    int GetNrows() const
    {
        return static_cast<int>(_nrows);
    }
    /** 	Get the number of rows.
    * 	@return number of rows.
    */

    int GetNcols() const
    {
        return static_cast<int>(_mcols);
    }

private:
    size_t _nrows;       //!< number of rows in matrix
    size_t _mcols;       //!< number of coluns in matrix
    std::vector<double> _ucol;  //!< column vector for tensor
    std::vector<double> _vrow;  //!< row vector for tensor

};


#endif
